﻿' Definitions:
' Project - a binary .seed file
' Active A - asset of type 19
' Active B - asset of type 21
' Active - either Active A or Active B
' Active describe - something like a definition of class with its fields. Actives A are then instances of such class.
' Active description - the set of all Active describes.

Public Module Seed
    Class Project
        Public Assets As New List(Of Asset)
        Private AssetLUT As Asset() 'For quick lookup by ID
        Public File As String

        Private Shared _activeDesc As ActiveDescription
        ReadOnly Property ActiveDescription As ActiveDescription
            Get
                If _activeDesc Is Nothing Then
                    Dim dir = IO.Path.GetDirectoryName(File)
                    Dim descFile = IO.Path.Combine(Dir, "Active description.seed")
                    If IO.File.Exists(descFile) Then
                        _activeDesc = New ActiveDescription(descFile)
                    Else
                        descFile = IO.Path.Combine(dir, "..\Active description.seed")
                        If IO.File.Exists(descFile) Then
                            _activeDesc = New ActiveDescription(descFile)
                        End If
                    End If
                End If
                If _activeDesc Is Nothing Then
                    Throw New Exception("Can't find ""Active description.seed""")
                End If
                Return _activeDesc
            End Get
        End Property

        Sub New(file As String)
            Me.File = file
            Using r As New IO.BinaryReader(New IO.FileStream(file, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.ReadWrite))
                If ReadFixedZString(r, 4) <> "bin " Then Throw New Exception("Bad file header")

                While r.BaseStream.Position < r.BaseStream.Length
                    Dim asset As Asset = Nothing

                    Dim offset = r.BaseStream.Position + 44
                    Dim size = r.ReadInt32
                    Dim id = r.ReadInt32
                    Dim type As AssetType = r.ReadInt32

                    Select Case type
                        Case AssetType.ActiveA : asset = New ActiveA
                        Case AssetType.ActiveB : asset = New ActiveB
                        Case AssetType.Light : asset = New Light
                        Case AssetType.Model : asset = New Model
                        Case AssetType.Room : asset = New Room
                        Case AssetType.Scene : asset = New Scene
                        Case AssetType.Texture : asset = New Texture
                        Case AssetType.Window : asset = New Window
                        Case Else : asset = New Asset
                    End Select

                    asset.Project = Me
                    asset.OriginalOffset = offset
                    asset.OriginalSize = size
                    asset.ID = id
                    asset.Type = type
                    asset.Name = ReadFixedZString(r, 32)
                    asset.Data = r.ReadBytes(size - 44)

                    Assets.Add(asset)
                End While

            End Using

            ' Calculate LUT for fast ID lookup
            Dim maxID = -1
            For Each a In Assets
                If a.ID > maxID Then maxID = a.ID
            Next
            ReDim AssetLUT(maxID)
            For Each Asset In Assets
                AssetLUT(Asset.ID) = Asset
            Next

            ' Actually parse stuff (we need to do this later, otherwise
            ' variables could point to not loaded yet assets).
            For Each a In Assets
                Using r As New IO.BinaryReader(New IO.MemoryStream(a.Data))
                    a.Deserialize(r)
                    If r.BaseStream.Position <> 0 AndAlso r.BaseStream.Position <> r.BaseStream.Length Then
                        Throw New Exception(a.Name & ": Serial size mismatch.")
                    End If
                End Using
            Next

            ' Find parents of each active.
            For Each a In Assets
                If a.Type = AssetType.ActiveB Then
                    Dim ab = DirectCast(a, ActiveB)
                    For Each child In ab.Children
                        If child.Parent IsNot Nothing Then
                            Throw New Exception("An Active can't have two parents.")
                        End If
                        child.Parent = a
                    Next
                End If
            Next

        End Sub

        Sub Save(file As String)
            Me.File = file
            Using w As New IO.BinaryWriter(New IO.FileStream(file, IO.FileMode.Create, IO.FileAccess.Write))
                w.Write(544106850)

                For Each a In Assets
                    If a.Data Is Nothing Then
                        Dim stream As New IO.MemoryStream()
                        Using w2 As New IO.BinaryWriter(stream)
                            a.Serialize(w2)
                        End Using
                        a.Data = stream.ToArray()
                        stream.Close()
                    End If

                    w.Write(a.Data.Length + 44)
                    w.Write(a.ID)
                    w.Write(a.Type)
                    WriteFixedString(w, a.Name, 32)
                    w.Write(a.Data)
                Next
            End Using
        End Sub

        Function GetAsset(id As Integer) As Asset
            If id = -1 Then Return Nothing
            Return AssetLUT(id)
        End Function

        Sub AllocateID(a As Asset)
            ' Find a free ID
            a.ID = -1
            For i = 0 To AssetLUT.Length - 1
                If AssetLUT(i) Is Nothing Then
                    a.ID = i
                    AssetLUT(i) = a
                    Exit For
                End If
            Next

            If a.ID = -1 Then
                ReDim Preserve AssetLUT(AssetLUT.Length)
                a.ID = AssetLUT.Length - 1
                AssetLUT(a.ID) = a
            End If
        End Sub

        Function CreateAsset(type As AssetType) As Asset
            Dim a As Asset = Nothing

            Select Case type
                Case AssetType.ActiveA : a = New ActiveA
                Case AssetType.ActiveB : a = New ActiveB
                Case AssetType.Light : a = New Light
                Case AssetType.Model : a = New Model
                Case AssetType.Room : a = New Room
                Case AssetType.Scene : a = New Scene
                Case AssetType.Texture : a = New Texture
                Case AssetType.Window : a = New Window
                Case Else : a = New Asset
            End Select
            a.Type = type

            a.Project = Me

            '' Find a free ID
            'a.ID = -1
            'For i = 0 To AssetLUT.Length - 1
            '    If AssetLUT(i) Is Nothing Then
            '        a.ID = i
            '        AssetLUT(i) = a
            '        Exit For
            '    End If
            'Next

            'If a.ID = -1 Then
            '    ReDim Preserve AssetLUT(AssetLUT.Length)
            '    a.ID = AssetLUT.Length - 1
            '    AssetLUT(a.ID) = a
            'End If
            AllocateID(a)

            Assets.Add(a)
            Return a
        End Function

        Sub RemoveAssets(which As Predicate(Of Asset))
            For i = 0 To AssetLUT.Length - 1
                If AssetLUT(i) IsNot Nothing AndAlso which(AssetLUT(i)) Then
                    AssetLUT(i) = Nothing
                End If
            Next

            Assets.RemoveAll(which)
        End Sub

        Function EnsureUniqueName(name As String) As String
            ' Trim out existing ~number
            Dim tildePos = name.IndexOf("~")
            If tildePos <> -1 Then
                Dim unused As Integer
                If Integer.TryParse(name.Substring(tildePos + 1), unused) Then
                    name = name.Substring(0, tildePos)
                End If
            End If

            Dim curName = name
            Dim i = 2
            While Assets.Find(Function(x) StrEq(x.Name, curName)) IsNot Nothing
                curName = name & "~" & i
                i += 1
            End While
            Return curName
        End Function
    End Class

    Class AnimFile
        Public Name As String
        Public Bones As New List(Of AnimBone)
        Public Sequences As New List(Of AnimSequence)

        Sub New(file As String)
            Using r As New IO.BinaryReader(New IO.FileStream(file, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.ReadWrite))
                While r.BaseStream.Position < r.BaseStream.Length
                    Dim chk As New AnimChunk
                    chk.Name = ReadFixedZString(r, 4)
                    chk.Offset = r.BaseStream.Position
                    Select Case chk.Name
                        Case "qNqN"
                            Name = ReadFixedZString(r, 64)
                            r.ReadInt32()
                            r.ReadInt32()
                            r.ReadInt32()
                        Case "tNqN"
                            Dim b As New AnimBone
                            r.ReadSingle()
                            b.Name = ReadFixedZString(r, 64)
                            b._parentName = ReadFixedZString(r, 64)
                            r.ReadInt32()
                            Bones.Add(b)
                        Case "rNqN"
                            Dim seq As New AnimSequence
                            seq.Name = ReadFixedZString(r, 64)
                            r.ReadSingle()
                            r.ReadInt32()
                            r.ReadInt32()
                            r.ReadInt32()
                            Sequences.Add(seq)
                        Case "sNqN"
                            Dim curSeq = Sequences(Sequences.Count - 1)
                            Dim curFrame As New AnimFrame
                            curFrame.Matrix = ReadMatrix(r)
                            r.ReadInt32()
                            r.ReadInt32()
                            curSeq.Frames.Add(curFrame)
                        Case "uNqN"
                            Dim curSeq = Sequences(Sequences.Count - 1)
                            Dim curFrame = curSeq.Frames(curSeq.Frames.Count - 1)
                            curFrame.BonePositions.Add(ReadMatrix(r))
                            r.ReadInt32()
                        Case Else
                            Throw New Exception("Unknown chunk " & chk.Name)
                    End Select
                End While
            End Using

            ' Set up the hiererchy.
            For Each bone In Bones
                Dim bone2 = bone
                If bone._parentName <> "" Then
                    bone.iParent = Bones.FindIndex(Function(x) x.Name = bone2._parentName)
                    If bone.iParent <> -1 Then
                        bone.Parent = Bones(bone.iParent)
                    End If
                End If
            Next
            For Each bone In Bones
                For Each bone2 In Bones
                    If bone2.Parent Is bone Then
                        bone.Children.Add(bone2)
                    End If
                Next
            Next

        End Sub

    End Class

    Structure AnimChunk
        Dim Name As String
        Dim Offset As Integer
        Dim Text As String
    End Structure

    Class AnimBone
        Public Name As String
        Public _parentName As String
        Public Parent As AnimBone
        Public iParent As Integer
        Public Children As New List(Of AnimBone)

        ' For use in editor
        Public CurMatrix As Matrix
        Public Selected As Boolean
    End Class

    Class AnimSequence
        Public Name As String
        Public Frames As New List(Of AnimFrame)
    End Class

    Class AnimFrame
        Public Matrix As Matrix
        Public BonePositions As New List(Of Matrix)
    End Class

    Enum AssetType
        Texture = 0
        Room = 3
        Window = 4
        Scene = 5
        Model = 17
        Light = 18
        ActiveA = 19
        ActiveB = 21
    End Enum

    Class Asset
        Implements IComparable(Of Asset)
        Public Project As Project
        Public OriginalOffset, OriginalSize As Integer
        Public ID As Integer
        Public Type As AssetType
        Public Name As String
        Public Data As Byte()

        Public Room As Room

        ' For use in editor
        Public Visible As Boolean = True
        Public Selected As Boolean

        Overridable Function GetLocation() As Vector
        End Function

        Overridable Sub AddLocation(delta As Vector)
        End Sub

        Protected Function GetAsset(id As Integer) As Asset
            Return Project.GetAsset(id)
        End Function

        Protected Function GetID(asset As Asset) As Integer
            If asset IsNot Nothing Then
                Return asset.ID
            Else
                Return -1
            End If
        End Function

        Overridable Sub Deserialize(r As IO.BinaryReader)
        End Sub

        Overridable Sub Serialize(w As IO.BinaryWriter)
        End Sub

        Sub EnsureType(type As AssetType)
            If Me.Type <> type Then
                Throw New Exception("Wrong asset type.")
            End If
        End Sub

        Function CompareTo(other As Asset) As Integer Implements IComparable(Of Asset).CompareTo
            Return ID.CompareTo(other.ID)
        End Function

        Overrides Function ToString() As String
            Return Type & "'" & Name & "'"
        End Function
    End Class

    Class Texture
        Inherits Asset
        Public Bitmap As Bitmap

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.Texture)

            Dim w = r.ReadInt32()
            Dim h = r.ReadInt32()

            Bitmap = New Bitmap(w, h)
            Dim fastBmp As New FastPixel(Bitmap)
            fastBmp.Lock()
            For y = 0 To h - 1
                For x = 0 To w - 1
                    Dim cb = r.ReadByte
                    Dim cg = r.ReadByte
                    Dim cr = r.ReadByte
                    fastBmp.SetPixel(x, y, Color.FromArgb(cr, cg, cb))
                Next
            Next
            fastBmp.Unlock(True)

            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.Texture)

            w.Write(Bitmap.Width)
            w.Write(Bitmap.Height)
            Dim fastBmp As New FastPixel(Bitmap)
            fastBmp.Lock()
            For y = 0 To Bitmap.Height - 1
                For x = 0 To Bitmap.Width - 1
                    Dim pix = fastBmp.GetPixel(x, y)
                    w.Write(pix.B)
                    w.Write(pix.G)
                    w.Write(pix.R)
                Next
            Next
            fastBmp.Unlock(False)
        End Sub

    End Class

    Class Room
        Inherits Asset
    End Class

    Enum ModelFlags
        Exists = 1
        FlatShaded = 16
        Translucent = 32
        NoCollision = 64
        Invisible1 = 128
        Invisible2 = 256
        NoShadowCast1 = 1024
        NoShadowCast2 = 2048
    End Enum

    Class Model
        Inherits Asset
        Public Unknown1 As Integer
        Public Matrix As Matrix
        Public PolygonGroups As New List(Of PolygonGroup)
        Public Polygons As New List(Of Polygon)
        Public Verts As New List(Of Vector)
        Public Textures As New List(Of Texture)
        Public Unknown3 As Integer
        Public Transparency As Single
        Public Flags As ModelFlags

        Public Overrides Function GetLocation() As Vector
            Return New Vector(Matrix.Xfrom1, Matrix.Yfrom1, Matrix.Zfrom1)
        End Function

        Public Overrides Sub AddLocation(delta As Vector)
            Matrix.Xfrom1 += delta.X
            Matrix.Yfrom1 += delta.Y
            Matrix.Zfrom1 += delta.Z
        End Sub

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.Model)

            Unknown1 = r.ReadInt32()
            Room = GetAsset(r.ReadInt32())

            'Matrix.Xfrom1 = r.ReadSingle()
            'Matrix.Yfrom1 = r.ReadSingle()
            'Matrix.Zfrom1 = r.ReadSingle()
            'Matrix.XfromX = r.ReadSingle()
            'Matrix.YfromX = r.ReadSingle()
            'Matrix.ZfromX = r.ReadSingle()
            'Matrix.XfromY = r.ReadSingle()
            'Matrix.YfromY = r.ReadSingle()
            'Matrix.ZfromY = r.ReadSingle()
            'Matrix.XfromZ = r.ReadSingle()
            'Matrix.YfromZ = r.ReadSingle()
            'Matrix.ZfromZ = r.ReadSingle()
            Matrix = ReadMatrix(r)

            Dim numUnknown = r.ReadInt32
            For i = 0 To numUnknown - 1
                Dim item As New PolygonGroup
                item.iFirstPolygon = r.ReadInt32()
                item.numPolygons = r.ReadInt32()
                PolygonGroups.Add(item)
            Next
            Dim numPolys = r.ReadInt32
            For i = 0 To numPolys - 1
                Dim p As New Polygon
                p.iTexture = r.ReadInt32()
                p.Diffuse = r.ReadSingle()
                p.EmissiveColor = ReadVector(r)
                p.Flags = r.ReadInt32()
                p.V1 = r.ReadInt32()
                p.V2 = r.ReadInt32()
                p.V3 = r.ReadInt32()
                p.V4 = r.ReadInt32()
                p.UV1.U = r.ReadSingle()
                p.UV1.V = r.ReadSingle()
                p.UV2.U = r.ReadSingle()
                p.UV2.V = r.ReadSingle()
                p.UV3.U = r.ReadSingle()
                p.UV3.V = r.ReadSingle()
                p.UV4.U = r.ReadSingle()
                p.UV4.V = r.ReadSingle()
                p.Unknown3 = r.ReadSingle()
                Polygons.Add(p)
            Next
            Dim numVerts = r.ReadInt32
            For i = 0 To numVerts - 1
                Verts.Add(ReadVector(r))
            Next
            Dim numTextures = r.ReadInt32
            For i = 0 To numTextures - 1
                Textures.Add(GetAsset(r.ReadInt32))
            Next
            Unknown3 = r.ReadInt32()
            Transparency = r.ReadSingle()
            Flags = r.ReadInt32()


            'test something
            Dim numUsedPolys As Integer
            For Each item In PolygonGroups
                If item.iFirstPolygon <> numUsedPolys Then
                    MessageBox.Show("item.iFirstPolygon <> numUsedPolys")
                End If
                numUsedPolys += item.numPolygons
            Next
            If numUsedPolys <> Polygons.Count Then
                MessageBox.Show("numUsedPolys <> result.Polygons.Count")
            End If

            ' Necessary for OBJ export
            Polygons.Sort(Function(x, y) x.iTexture.CompareTo(y.iTexture))

            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.Model)

            w.Write(Unknown1)
            w.Write(GetID(Room))
            w.Write(Matrix.Xfrom1)
            w.Write(Matrix.Yfrom1)
            w.Write(Matrix.Zfrom1)
            w.Write(Matrix.XfromX)
            w.Write(Matrix.YfromX)
            w.Write(Matrix.ZfromX)
            w.Write(Matrix.XfromY)
            w.Write(Matrix.YfromY)
            w.Write(Matrix.ZfromY)
            w.Write(Matrix.XfromZ)
            w.Write(Matrix.YfromZ)
            w.Write(Matrix.ZfromZ)

            w.Write(PolygonGroups.Count)
            For Each item In PolygonGroups
                w.Write(item.iFirstPolygon)
                w.Write(item.numPolygons)
            Next

            w.Write(Polygons.Count)
            For Each poly In Polygons
                w.Write(poly.iTexture)
                w.Write(poly.Diffuse)
                WriteVector(w, poly.EmissiveColor)
                w.Write(poly.Flags)
                w.Write(poly.V1)
                w.Write(poly.V2)
                w.Write(poly.V3)
                w.Write(poly.V4)
                w.Write(poly.UV1.U)
                w.Write(poly.UV1.V)
                w.Write(poly.UV2.U)
                w.Write(poly.UV2.V)
                w.Write(poly.UV3.U)
                w.Write(poly.UV3.V)
                w.Write(poly.UV4.U)
                w.Write(poly.UV4.V)
                w.Write(poly.Unknown3)
            Next

            w.Write(Verts.Count)
            For Each vert In Verts
                WriteVector(w, vert)
            Next

            w.Write(Textures.Count)
            For Each tex In Textures
                w.Write(GetID(tex))
            Next

            w.Write(Unknown3)
            w.Write(Transparency)
            w.Write(Flags)
        End Sub

        Sub New()
            Matrix.XfromX = 1
            Matrix.YfromY = 1
            Matrix.ZfromZ = 1
            Flags = ModelFlags.Exists
        End Sub

        Sub LoadOBJ(file As String)
            Dim UVs As New List(Of UV)
            Dim cur_iTexture As Integer = -1

            Using r As New IO.StreamReader(file)
                While Not r.EndOfStream
                    Dim line = r.ReadLine
                    If line.StartsWith("v ") Then
                        Dim parts = line.Split({" "}, StringSplitOptions.RemoveEmptyEntries)
                        Dim v As New Vector
                        v.X = StrToNum(parts(1))
                        v.Y = StrToNum(parts(2))
                        v.Z = StrToNum(parts(3))
                        Verts.Add(v)
                    ElseIf line.StartsWith("vt ") Then
                        Dim parts = line.Split({" "}, StringSplitOptions.RemoveEmptyEntries)
                        Dim uv As New UV
                        uv.U = StrToNum(parts(1))
                        uv.V = StrToNum(parts(2))
                        UVs.Add(uv)
                    ElseIf line.StartsWith("f ") Then
                        Dim parts = line.Split({" "}, StringSplitOptions.RemoveEmptyEntries)
                        If parts.Length <> 4 And parts.Length <> 5 Then
                            Throw New Exception("Model must have tris or quads. (FIXME)")
                        End If
                        Dim poly As New Polygon
                        For i = 1 To parts.Length - 1
                            Dim subparts = parts(i).Split("/")
                            poly.V(i) = subparts(0) - 1
                            If subparts(1) <> "" Then
                                poly.UV(i) = UVs(subparts(1) - 1)
                            End If
                        Next
                        If parts.Length = 4 Then
                            poly.V4 = poly.V3
                            poly.UV4 = poly.UV3
                        End If
                        poly.iTexture = cur_iTexture
                        poly.Diffuse = 1
                        poly.EmissiveColor = New Vector(0, 0, 0)
                        poly.Flags = 255 'IDK
                        Polygons.Add(poly)
                    ElseIf line.StartsWith("usemtl ") Then
                        Dim texName = line.Substring(7)
                        Dim texAsset As Texture = Project.Assets.Find(Function(x) x.Type = AssetType.Texture And StrEq(x.Name, texName))
                        If texAsset Is Nothing Then
                            MessageBox.Show("Bad texture name (" & texName & ")")
                            cur_iTexture = -1
                        Else
                            Textures.Add(texAsset)
                            cur_iTexture = Textures.Count - 1
                        End If
                    End If
                End While
            End Using
            PolygonGroups.Add(New PolygonGroup(0, Polygons.Count)) ' ???
        End Sub

        Sub SaveOBJ(file As String)
            Using w As New IO.StreamWriter(file)
                w.NewLine = Chr(10)
                For Each v In Verts
                    Dim trans_v = v.TransformBy(Matrix)
                    w.WriteLine("v " & NumToStr(trans_v.X) & " " & NumToStr(trans_v.Y) & " " & NumToStr(trans_v.Z))
                Next
                For Each p In Polygons
                    For iV = 1 To 4
                        w.WriteLine("vt " & NumToStr(p.UV(iV).U) & " " & NumToStr(1 - p.UV(iV).V))
                    Next
                Next

                Dim prev_iTexture As Integer
                Dim first = True
                Dim curUV = 1
                For Each p In Polygons
                    If first OrElse p.iTexture <> prev_iTexture Then
                        Dim p2 = p
                        Dim newTexName As String
                        If p2.iTexture >= 0 Then
                            'newTexName = Me.Project.Assets.Find(Function(x) x.ID = Textures(p2.iTexture)).Name
                            newTexName = Textures(p2.iTexture).Name
                        Else
                            '???
                            newTexName = "(Bad tex slot " & p2.iTexture & ")"
                        End If
                        w.WriteLine("g " & newTexName)
                        w.WriteLine("usemtl " & newTexName)
                    End If

                    w.Write("f")
                    w.Write(" " & (p.V1 + 1) & "/" & curUV)
                    w.Write(" " & (p.V2 + 1) & "/" & curUV + 1)
                    w.Write(" " & (p.V3 + 1) & "/" & curUV + 2)
                    w.Write(" " & (p.V4 + 1) & "/" & curUV + 3)
                    w.WriteLine()

                    curUV += 4
                    prev_iTexture = p.iTexture
                    first = False
                Next

            End Using
        End Sub

        Function Clone() As Model
            Dim copy As Model = Me.MemberwiseClone()
            copy.Data = {}

            copy.PolygonGroups = New List(Of PolygonGroup)
            copy.Polygons = New List(Of Polygon)
            copy.Textures = New List(Of Texture)
            copy.Verts = New List(Of Vector)

            copy.PolygonGroups.AddRange(PolygonGroups)
            copy.Polygons.AddRange(Polygons)
            copy.Textures.AddRange(Textures)
            copy.Verts.AddRange(Verts)

            Return copy
        End Function

    End Class

    Structure PolygonGroup
        Dim iFirstPolygon, numPolygons As Integer

        Sub New(iFirst As Integer, num As Integer)
            iFirstPolygon = iFirst
            numPolygons = num
        End Sub
    End Structure

    Structure Polygon
        Dim iTexture As Integer
        Dim Diffuse As Single
        Dim EmissiveColor As Vector
        Dim Flags As Integer
        Dim V1, V2, V3, V4 As Integer
        Dim UV1, UV2, UV3, UV4 As UV
        Dim Unknown3 As Single

        Property V(num As Integer) As Integer
            Get
                Select Case num
                    Case 1 : Return V1
                    Case 2 : Return V2
                    Case 3 : Return V3
                    Case 4 : Return V4
                    Case Else
                        Throw New Exception("Bad V() num")
                End Select
            End Get
            Set(value As Integer)
                Select Case num
                    Case 1 : V1 = value
                    Case 2 : V2 = value
                    Case 3 : V3 = value
                    Case 4 : V4 = value
                    Case Else
                        Throw New Exception("Bad V() num")
                End Select
            End Set
        End Property

        Property UV(num As Integer) As UV
            Get
                Select Case num
                    Case 1 : Return UV1
                    Case 2 : Return UV2
                    Case 3 : Return UV3
                    Case 4 : Return UV4
                    Case Else
                        Throw New Exception("Bad UV() num")
                End Select
            End Get
            Set(value As UV)
                Select Case num
                    Case 1 : UV1 = value
                    Case 2 : UV2 = value
                    Case 3 : UV3 = value
                    Case 4 : UV4 = value
                    Case Else
                        Throw New Exception("Bad UV() num")
                End Select
            End Set
        End Property


    End Structure

    Structure UV
        Dim U, V As Single
    End Structure

    Structure Vector
        Dim X, Y, Z As Single

        Sub New(x As Single, y As Single, z As Single)
            Me.X = x
            Me.Y = y
            Me.Z = z
        End Sub

        Function TransformBy(mtx As Matrix) As Vector
            Dim o As Vector
            o.X = mtx.Xfrom1 + X * mtx.XfromX + Y * mtx.XfromY + Z * mtx.XfromZ
            o.Y = mtx.Yfrom1 + X * mtx.YfromX + Y * mtx.YfromY + Z * mtx.YfromZ
            o.Z = mtx.Zfrom1 + X * mtx.ZfromX + Y * mtx.ZfromY + Z * mtx.ZfromZ
            Return o
        End Function

        Function ToUnreal() As Vector
            Dim o As Vector
            o.X = X
            o.Y = -Z
            o.Z = Y
            Return o
        End Function

        Shared Operator +(lhs As Vector, rhs As Vector) As Vector
            Return New Vector(lhs.X + rhs.X, lhs.Y + rhs.Y, lhs.Z + rhs.Z)
        End Operator

        Shared Operator -(lhs As Vector, rhs As Vector) As Vector
            Return New Vector(lhs.X - rhs.X, lhs.Y - rhs.Y, lhs.Z - rhs.Z)
        End Operator

        Shared Operator *(lhs As Vector, rhs As Single) As Vector
            Return New Vector(lhs.X * rhs, lhs.Y * rhs, lhs.Z * rhs)
        End Operator

        Shared Operator *(lhs As Single, rhs As Vector) As Vector
            Return New Vector(rhs.X * lhs, rhs.Y * lhs, rhs.Z * lhs)
        End Operator

        ''' <summary>
        ''' Component by component multiplication.
        ''' </summary>
        Shared Operator *(lhs As Vector, rhs As Vector) As Vector
            Return New Vector(rhs.X * lhs.X, rhs.Y * lhs.Y, rhs.Z * lhs.Z)
        End Operator

        Shared Operator /(lhs As Vector, rhs As Single) As Vector
            Return New Vector(lhs.X / rhs, lhs.Y / rhs, lhs.Z / rhs)
        End Operator

        Shared Function Cross(v1 As Vector, v2 As Vector) As Vector
            Return New Vector(v1.Y * v2.Z - v1.Z * v2.Y, v1.Z * v2.X - v1.X * v2.Z, v1.X * v2.Y - v1.Y * v2.X)
        End Function

        Function Normalize() As Vector
            Dim len = Math.Sqrt(X * X + Y * Y + Z * Z)
            Return New Vector(X / len, Y / len, Z / len)
        End Function
    End Structure

    Structure Matrix
        Dim XfromX, XfromY, XfromZ, Xfrom1 As Single
        Dim YfromX, YfromY, YfromZ, Yfrom1 As Single
        Dim ZfromX, ZfromY, ZfromZ, Zfrom1 As Single

        Sub New(i0 As Single, i1 As Single, i2 As Single, i3 As Single, i4 As Single, i5 As Single, i6 As Single, i7 As Single, i8 As Single, i9 As Single, i10 As Single, i11 As Single)
            Item(0) = i0
            Item(1) = i1
            Item(2) = i2
            Item(3) = i3
            Item(4) = i4
            Item(5) = i5
            Item(6) = i6
            Item(7) = i7
            Item(8) = i8
            Item(9) = i9
            Item(10) = i10
            Item(11) = i11
        End Sub

        Default Property Item(i As Integer) As Single
            Get
                Select Case i
                    Case 0 : Return XfromX
                    Case 1 : Return YfromX
                    Case 2 : Return ZfromX
                    Case 3 : Return XfromY
                    Case 4 : Return YfromY
                    Case 5 : Return ZfromY
                    Case 6 : Return XfromZ
                    Case 7 : Return YfromZ
                    Case 8 : Return ZfromZ
                    Case 9 : Return Xfrom1
                    Case 10 : Return Yfrom1
                    Case 11 : Return Zfrom1
                End Select
                Throw New IndexOutOfRangeException
            End Get
            Set(value As Single)
                If i < 0 Or i > 11 Then Throw New IndexOutOfRangeException
                Select Case i
                    Case 0 : XfromX = value
                    Case 1 : YfromX = value
                    Case 2 : ZfromX = value
                    Case 3 : XfromY = value
                    Case 4 : YfromY = value
                    Case 5 : ZfromY = value
                    Case 6 : XfromZ = value
                    Case 7 : YfromZ = value
                    Case 8 : ZfromZ = value
                    Case 9 : Xfrom1 = value
                    Case 10 : Yfrom1 = value
                    Case 11 : Zfrom1 = value
                End Select
            End Set
        End Property

        Default Property Item(row As Integer, col As Integer) As Single
            Get
                If row = 3 Then
                    Return If(col = 3, 1, 0)
                End If
                Return Item(row + col * 3)
            End Get
            Set(value As Single)
                Item(row + col * 3) = value
            End Set
        End Property

        Shared Operator *(lhs As Matrix, rhs As Matrix) As Matrix
            Dim result As Matrix

            For outRow = 0 To 2
                For outCol = 0 To 3
                    For i = 0 To 3
                        result(outRow, outCol) += lhs(outRow, i) * rhs(i, outCol)
                    Next
                Next
            Next

            Return result
        End Operator

        Shared ReadOnly Property Identity As Matrix
            Get
                Return New Matrix(1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0)
            End Get
        End Property

        Shared ReadOnly Property Rot90 As Matrix
            Get
                Return New Matrix(0, 0, -1, 0, 1, 0, 1, 0, 0, 0, 0, 0)
            End Get
        End Property

        Shared ReadOnly Property RotNeg90 As Matrix
            Get
                Return New Matrix(0, 0, 1, 0, 1, 0, -1, 0, 0, 0, 0, 0)
            End Get
        End Property

        Shared ReadOnly Property Rot45 As Matrix
            Get
                Dim s = Math.Sqrt(2) / 2
                Return New Matrix(s, 0, -s, 0, 1, 0, s, 0, s, 0, 0, 0)
            End Get
        End Property

        Shared ReadOnly Property RotNeg45 As Matrix
            Get
                Dim s = Math.Sqrt(2) / 2
                Return New Matrix(s, 0, s, 0, 1, 0, -s, 0, s, 0, 0, 0)
            End Get
        End Property

        Shared Function RotationX(degrees As Single) As Matrix
            Dim rad = degrees / 360 * 2 * Math.PI
            Return New Matrix(1, 0, 0, 0, Math.Cos(rad), -Math.Sin(rad), 0, Math.Sin(rad), Math.Cos(rad), 0, 0, 0)
        End Function

        Shared Function RotationY(degrees As Single) As Matrix
            Dim rad = degrees / 360 * 2 * Math.PI
            Return New Matrix(Math.Cos(rad), 0, -Math.Sin(rad), 0, 1, 0, Math.Sin(rad), 0, Math.Cos(rad), 0, 0, 0)
        End Function

        Shared Function RotationZ(degrees As Single) As Matrix
            Dim rad = degrees / 360 * 2 * Math.PI
            Return New Matrix(Math.Cos(rad), -Math.Sin(rad), 0, Math.Sin(rad), Math.Cos(rad), 0, 0, 0, 1, 0, 0, 0)
        End Function

        Shared Function Scale(v As Vector) As Matrix
            Return New Matrix(v.X, 0, 0, 0, v.Y, 0, 0, 0, v.Z, 0, 0, 0)
        End Function

        Shared Function Translation(v As Vector) As Matrix
            Return New Matrix(1, 0, 0, 0, 1, 0, 0, 0, 1, v.X, v.Y, v.Z)
        End Function

    End Structure

    Class Light
        Inherits Asset
        Public Location As Vector
        Public Radius As Single = 512
        Public Brightness As Single = 2
        Public Color As Vector = New Vector(255, 255, 255)
        Public Unknown1 As Integer
        Public Unknown2 As Integer = 1

        Public Overrides Function GetLocation() As Vector
            Return Location
        End Function

        Public Overrides Sub AddLocation(delta As Vector)
            Location += delta
        End Sub

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.Light)

            Room = GetAsset(r.ReadInt32())
            Location = ReadVector(r)
            Radius = r.ReadSingle()
            Brightness = r.ReadSingle()
            Color = ReadVector(r)
            Unknown1 = r.ReadInt32()
            Unknown2 = r.ReadInt32()

            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.Light)

            w.Write(GetID(Room))
            WriteVector(w, Location)
            w.Write(Radius)
            w.Write(Brightness)
            WriteVector(w, Color)
            w.Write(Unknown1)
            w.Write(Unknown2)
        End Sub

        Function Clone() As Light
            Dim copy = Me.MemberwiseClone()
            copy.Data = {}
            Return copy
        End Function

    End Class

    Class Window
        Inherits Asset
        Public Room1, Room2 As Room
        Public Verts As New List(Of Vector)

        Public Overrides Function GetLocation() As Vector
            Dim sum As Vector
            For Each v In Verts
                sum += v
            Next
            Return sum / Verts.Count
        End Function

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.Window)

            Room1 = GetAsset(r.ReadInt32)
            Room2 = GetAsset(r.ReadInt32)
            Dim numVerts = r.ReadInt32()
            For i = 0 To numVerts - 1
                Verts.Add(ReadVector(r))
            Next

            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.Window)

            w.Write(GetID(Room1))
            w.Write(GetID(Room2))
            w.Write(Verts.Count)
            For Each v In Verts
                WriteVector(w, v)
            Next
        End Sub
    End Class

    Class Scene
        Inherits Asset
        Public Rooms As New List(Of Room)
        Public Windows As New List(Of Window)
        Public Models As New List(Of Model)
        Public Lights As New List(Of Light)
        Public Actives As New List(Of Active)

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.Scene)

            If r.BaseStream.Length <> 0 Then
                Dim numRooms = r.ReadInt32
                For i = 0 To numRooms - 1
                    Rooms.Add(GetAsset(r.ReadInt32))
                Next
                Dim numWindows = r.ReadInt32
                For i = 0 To numWindows - 1
                    Windows.Add(GetAsset(r.ReadInt32))
                Next
                Dim numModels = r.ReadInt32
                For i = 0 To numModels - 1
                    Models.Add(GetAsset(r.ReadInt32))
                Next
                Dim numLights = r.ReadInt32
                For i = 0 To numLights - 1
                    Lights.Add(GetAsset(r.ReadInt32))
                Next
                Dim numActives = r.ReadInt32
                For i = 0 To numActives - 1
                    Actives.Add(GetAsset(r.ReadInt32))
                Next
            End If

            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.Scene)

            w.Write(Rooms.Count)
            For Each rm In Rooms
                w.Write(GetID(rm))
            Next
            w.Write(Windows.Count)
            For Each window In Windows
                w.Write(GetID(window))
            Next
            w.Write(Models.Count)
            For Each model In Models
                w.Write(GetID(model))
            Next
            w.Write(Lights.Count)
            For Each light In Lights
                w.Write(GetID(light))
            Next
            w.Write(Actives.Count)
            For Each active In Actives
                w.Write(GetID(active))
            Next
        End Sub
    End Class

    MustInherit Class Active
        Inherits Asset
        Public Location As Vector
        Public Parent As ActiveB

        Public Overrides Function GetLocation() As Vector
            Return Location
        End Function

        Public Overrides Sub AddLocation(delta As Vector)
            Location += delta
        End Sub

        MustOverride Function GetNumParams() As Integer
        MustOverride Function GetParamName(iParam As Integer) As String
        MustOverride Function GetParam(iParam As Integer) As ActiveParam
    End Class

    Class ActiveA
        Inherits Active
        Public Unknown1 As Vector
        Public Unknown2 As Vector
        Public Unknown3 As Single
        Public Describe As ActiveDescribe
        Public Unknown4 As Integer
        Public Params As New List(Of ActiveParam)

        Function GetCodeString() As String
            Dim result As New System.Text.StringBuilder

            result.Append(Describe.Name & "(")

            Dim first = True
            For Each param In Params
                If Not first Then
                    result.Append(", ")
                End If
                result.Append(param.ToString())
                first = False
            Next

            result.Append(")")

            Return result.ToString
        End Function

        'Default ReadOnly Property GetParamValue(iParam As Integer) As String
        '    Get
        '        Return Params(iParam).Value
        '    End Get
        'End Property

        'Default ReadOnly Property GetParamValue(name As String) As String
        '    Get
        '        For Each p In Params
        '            If p.Definition.Name = name Then
        '                Return p.Value
        '            End If
        '        Next
        '        Return ""
        '    End Get
        'End Property

        Public Overrides Function GetNumParams() As Integer
            Return Params.Count
        End Function

        Public Overrides Function GetParam(iParam As Integer) As ActiveParam
            Return Params(iParam)
        End Function

        Public Overrides Function GetParamName(iParam As Integer) As String
            Return Params(iParam).Definition.Name
        End Function

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.ActiveA)

            Room = GetAsset(r.ReadInt32())
            Location = ReadVector(r)
            Unknown1 = ReadVector(r)
            Unknown2 = ReadVector(r)
            Unknown3 = r.ReadSingle()
            Dim DescribeID = r.ReadInt32
            Unknown4 = r.ReadInt32()
            Dim numProps = r.ReadInt32

            Describe = Project.ActiveDescription().Chunks.Find(Function(x) x.ID = DescribeID)
            If Describe.ParameterLines.Count <> numProps Then
                Throw New Exception("numProps mismatch")
            End If

            For Each param In Describe.ParameterLines
                Params.Add(param.Parse(r, Project))
            Next


            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.ActiveA)

            w.Write(GetID(Room))
            WriteVector(w, Location)
            WriteVector(w, Unknown1)
            WriteVector(w, Unknown2)
            w.Write(Unknown3)
            w.Write(Describe.ID)
            w.Write(Unknown4)

            w.Write(Params.Count)
            For Each param In Params
                Select Case param.Definition.Type
                    Case ParamType.PT_Flag, ParamType.PT_List
                        w.Write(CInt(param.Value))
                    Case ParamType.PT_Float
                        w.Write(StrToNum(param.Value))
                    Case ParamType.PT_String
                        w.Write(param.Value.Length)
                        For i = 0 To param.Value.Length - 1
                            w.Write(CByte(Asc(param.Value(i))))
                        Next
                End Select
            Next
        End Sub

    End Class

    Class ActiveB
        Inherits Active
        Public Unknown1 As Vector
        Public Unknown2 As Vector
        Public Unknown3 As Single
        Public Unknown4 As Integer
        Public Unknown5 As Integer
        Public Active3 As Active
        Public NumSomething As Integer
        Public Active1 As Active
        Public Active2 As Active
        Public Children As New List(Of Active)
        Public ExposedParams As New List(Of ExposedParam)

        ' For use in editor
        Public PrefabName As String

        Public Overrides Function GetNumParams() As Integer
            Return ExposedParams.Count
        End Function

        Public Overrides Function GetParam(iParam As Integer) As ActiveParam
            Dim exp = ExposedParams(iParam)
            Return exp.Active.GetParam(exp.iParam)
        End Function

        Public Overrides Function GetParamName(iParam As Integer) As String
            Return ExposedParams(iParam).Name
        End Function

        Public Overrides Sub Deserialize(r As IO.BinaryReader)
            EnsureType(AssetType.ActiveB)

            Room = GetAsset(r.ReadInt32())
            Location = ReadVector(r)
            Unknown1 = ReadVector(r)
            Unknown2 = ReadVector(r)
            Unknown3 = r.ReadSingle()
            Unknown4 = r.ReadInt32()
            If Unknown4 <> -1 Then
                If True Then
                End If
            End If
            NumSomething = r.ReadInt32()
            Unknown5 = r.ReadInt32()
            If Unknown5 <> 0 Then
                If True Then
                End If
            End If
            Active1 = GetAsset(r.ReadInt32())
            Active2 = GetAsset(r.ReadInt32())
            Active3 = GetAsset(r.ReadInt32())
            Dim numChildren = r.ReadInt32()
            For i = 0 To numChildren - 1
                Children.Add(GetAsset(r.ReadInt32()))
            Next
            Dim count2 = r.ReadInt32()
            For i = 0 To count2 - 1
                Dim item As New ExposedParam
                item.Active = GetAsset(r.ReadInt32())
                item.iParam = r.ReadInt32()
                item.Unknown = r.ReadInt32()
                item.Name = ReadFixedZString(r, 36)
                ExposedParams.Add(item)
            Next

            ' Something similar to ActiveA's Describe.Name
            Dim tildePos = Name.IndexOf("~")
            If tildePos = -1 Then
                PrefabName = Name
            Else
                PrefabName = Name.Substring(0, tildePos)
            End If

            Data = Nothing
        End Sub

        Public Overrides Sub Serialize(w As IO.BinaryWriter)
            EnsureType(AssetType.ActiveB)

            w.Write(GetID(Room))
            WriteVector(w, Location)
            WriteVector(w, Unknown1)
            WriteVector(w, Unknown2)
            w.Write(Unknown3)
            w.Write(Unknown4)
            w.Write(NumSomething)
            w.Write(Unknown5)
            w.Write(GetID(Active1))
            w.Write(GetID(Active2))
            w.Write(GetID(Active3))
            w.Write(Children.Count)
            For Each c In Children
                w.Write(GetID(c))
            Next
            w.Write(ExposedParams.Count)
            For Each item In ExposedParams
                w.Write(GetID(item.Active))
                w.Write(item.iParam)
                w.Write(item.Unknown)
                WriteFixedString(w, item.Name, 36)
            Next
        End Sub
    End Class

    Structure ExposedParam
        Dim Active As Active
        Dim iParam As Integer
        Dim Unknown As Integer
        Dim Name As String
    End Structure

    Class ActiveParam
        Public Definition As ParameterLine
        Public Value As String = ""

        Overrides Function ToString() As String
            Select Case Definition.Type
                Case ParamType.PT_String
                    Return """" & Value.Replace("""", """""") & """"
                Case ParamType.PT_Float
                    Return Value
                Case ParamType.PT_List
                    Return Definition.Values(CInt(Value))
                Case ParamType.PT_Flag
                    Dim result = ""
                    Dim data = CInt(Value)
                    Dim first = True
                    For i = 0 To Definition.Values.Count - 1
                        If data And (1 << i) Then
                            If Not first Then
                                result &= "|"
                            End If
                            result &= Definition.Values(i)
                            first = False
                        End If
                    Next
                    Return result
            End Select
            Return Value
        End Function
    End Class

    Class ActiveDescription
        Public Chunks As New List(Of ActiveDescribe)

        Sub New(file As String)
            Using r As New IO.StreamReader(file, System.Text.Encoding.GetEncoding(1250))
                Dim curChunk As ActiveDescribe = Nothing
                Dim parsingParams As Boolean
                Dim lastParam As ParameterLine = Nothing

                While Not r.EndOfStream
                    Dim line = r.ReadLine
                    If line <> "" Then
                        If line = "[chunk]" Then
                            curChunk = New ActiveDescribe
                            Chunks.Add(curChunk)
                            parsingParams = False
                            lastParam = Nothing
                        ElseIf line.Contains("=") Then
                            Dim equalsPos = line.IndexOf("=")
                            Dim prop = line.Substring(0, equalsPos)
                            Dim value = line.Substring(equalsPos + 1)

                            If Not parsingParams Then
                                Select Case prop
                                    Case "id"
                                        curChunk.ID = value
                                    Case "name"
                                        curChunk.Name = value
                                    Case "parameterline"
                                        parsingParams = True
                                End Select
                            Else
                                lastParam = New ParameterLine
                                curChunk.ParameterLines.Add(lastParam)
                                lastParam.Name = prop
                                Dim parts = value.Split(",")
                                Dim typeStr = parts(0)
                                Dim ListCount = 0
                                Select Case typeStr
                                    Case "string" : lastParam.Type = ParamType.PT_String
                                    Case "float" : lastParam.Type = ParamType.PT_Float
                                    Case "list" : lastParam.Type = ParamType.PT_List
                                    Case "flag" : lastParam.Type = ParamType.PT_Flag
                                End Select
                                If parts.Length >= 2 Then
                                    lastParam.Misc = parts(1)
                                End If
                                If parts.Length >= 3 AndAlso parts(2) <> "" Then
                                    ListCount = parts(2)
                                End If
                            End If
                        ElseIf line.StartsWith("  ") Then
                            If lastParam.Type = ParamType.PT_List Or lastParam.Type = ParamType.PT_Flag Then
                                lastParam.Values.Add(line.Substring(2))
                            Else
                                Throw New Exception("Unexpected list element")
                            End If
                        End If
                    End If
                End While
            End Using
        End Sub
    End Class

    Class ActiveDescribe
        Public ID As Integer
        Public Name As String
        Public ParameterLines As New List(Of ParameterLine)
    End Class

    Class ParameterLine
        Public Name As String
        Public Type As ParamType
        Public Misc As String
        Public Values As New List(Of String)

        Function Parse(r As IO.BinaryReader, project As Project) As ActiveParam
            Dim result As New ActiveParam
            result.Definition = Me
            Select Case Type
                Case ParamType.PT_String
                    Dim len = r.ReadInt32
                    Dim text = ReadFixedZString(r, len)
                    result.Value = text
                Case ParamType.PT_Float
                    result.Value = NumToStr(r.ReadSingle)
                Case ParamType.PT_List, ParamType.PT_Flag
                    result.Value = r.ReadInt32
            End Select
            Return result
        End Function
    End Class

    Enum ParamType
        PT_String
        PT_Float
        PT_List
        PT_Flag
    End Enum

    Class CustomActiveExporter
        Dim w As IO.StreamWriter
        Dim indent As Integer
        Dim donePrefabs As New List(Of String)

        Private Sub StartLine(text As String)
            For i = 0 To indent - 1
                w.Write(vbTab)
            Next
            w.Write(text)
        End Sub

        Private Sub WriteLine(text As String)
            For i = 0 To indent - 1
                w.Write(vbTab)
            Next
            w.WriteLine(text)
        End Sub

        Function GetID(a As Asset) As Integer
            If a Is Nothing Then
                Return -1
            Else
                Return a.ID
            End If
        End Function

        Private Sub Export(act As ActiveA)
            WriteLine("default $" & act.ID & " " & act.GetCodeString() & " " & act.Name & " " & act.Unknown4)
        End Sub

        Private Sub Export(act As ActiveB)
            StartLine("custom $" & act.ID & " " & act.Name)
            w.Write(" " & NumToStr(act.Location.X) & " " & NumToStr(act.Location.Y) & " " & NumToStr(act.Location.Z))
            w.Write(" " & NumToStr(act.Unknown1.X) & " " & NumToStr(act.Unknown1.Y) & " " & NumToStr(act.Unknown1.Z))
            w.Write(" " & NumToStr(act.Unknown2.X) & " " & NumToStr(act.Unknown2.Y) & " " & NumToStr(act.Unknown2.Z))
            w.Write(" " & NumToStr(act.Unknown3))
            w.Write(" " & act.Unknown4)
            w.Write(" " & act.Unknown5)
            w.Write(" " & act.NumSomething)
            w.WriteLine()
            indent += 1
            WriteLine("Misc:")
            indent += 1
            WriteLine("$" & GetID(act.Active1))
            WriteLine("$" & GetID(act.Active2))
            WriteLine("$" & GetID(act.Active3))
            indent -= 1

            WriteLine("Children:")
            indent += 1
            For Each child In act.Children
                If TypeOf child Is ActiveA Then
                    Export(DirectCast(child, ActiveA))
                Else
                    Export(DirectCast(child, ActiveB))
                End If
            Next
            indent -= 1

            WriteLine("Exposed:")
            indent += 1
            For Each exp In act.ExposedParams
                WriteLine("$" & exp.Active.ID & "." & exp.iParam & " as " & exp.Name & " " & exp.Unknown)
            Next
            indent -= 1

            indent -= 1
        End Sub

        Private Sub ExportPrefab(act As ActiveB)
            If donePrefabs.Contains(act.PrefabName) Then Return
            WriteLine("[" & act.PrefabName & "]")
            indent += 1
            Export(act)
            indent -= 1
            WriteLine("")
            donePrefabs.Add(act.PrefabName)
        End Sub

        Sub Export(acts As List(Of ActiveB), file As String)
            w = New IO.StreamWriter(file)

            For Each act In acts
                ExportPrefab(act)
            Next

            w.Close()
            donePrefabs.Clear()
        End Sub
    End Class
End Module
